@@ -50,10 +50,9 @@ import (
 	"github.com/prometheus/common/model"
 	"github.com/prometheus/common/route"
 	"github.com/prometheus/common/server"
-	"github.com/soheilhy/cmux"
+	"github.com/prometheus/exporter-toolkit/https"
 	"go.uber.org/atomic"
 	"golang.org/x/net/netutil"
-	"google.golang.org/grpc"
 
 	"github.com/prometheus/prometheus/config"
 	"github.com/prometheus/prometheus/notifier"
@@ -66,13 +65,11 @@ import (
 	"github.com/prometheus/prometheus/tsdb/index"
 	"github.com/prometheus/prometheus/util/httputil"
 	api_v1 "github.com/prometheus/prometheus/web/api/v1"
-	api_v2 "github.com/prometheus/prometheus/web/api/v2"
 	"github.com/prometheus/prometheus/web/ui"
 )
 
 // Paths that are handled by the React / Reach router that should all be served the main React app's index.html.
 var reactRouterPaths = []string{
-	"/",
 	"/alerts",
 	"/config",
 	"/flags",
@@ -82,7 +79,6 @@ var reactRouterPaths = []string{
 	"/status",
 	"/targets",
 	"/tsdb-status",
-	"/version",
 }
 
 // withStackTrace logs the stack trace in case the request panics. The function
@@ -191,6 +187,7 @@ type Handler struct {
 
 	router      *route.Router
 	quitCh      chan struct{}
+	quitOnce    sync.Once
 	reloadCh    chan chan error
 	options     *Options
 	config      *config.Config
@@ -324,6 +321,7 @@ func New(logger log.Logger, o *Options) *Handler {
 		h.options.CORSOrigin,
 		h.runtimeInfo,
 		h.versionInfo,
+		o.Gatherer,
 	)
 
 	if o.RoutePrefix != "/" {
@@ -339,17 +337,40 @@ func New(logger log.Logger, o *Options) *Handler {
 	router.Get("/", func(w http.ResponseWriter, r *http.Request) {
 		http.Redirect(w, r, path.Join(o.ExternalURL.Path, "/graph"), http.StatusFound)
 	})
+	router.Get("/classic/", func(w http.ResponseWriter, r *http.Request) {
+		http.Redirect(w, r, path.Join(o.ExternalURL.Path, "/classic/graph"), http.StatusFound)
+	})
 
-	router.Get("/alerts", readyf(h.alerts))
-	router.Get("/graph", readyf(h.graph))
-	router.Get("/status", readyf(h.status))
-	router.Get("/flags", readyf(h.flags))
-	router.Get("/config", readyf(h.serveConfig))
-	router.Get("/rules", readyf(h.rules))
-	router.Get("/targets", readyf(h.targets))
-	router.Get("/version", readyf(h.version))
-	router.Get("/service-discovery", readyf(h.serviceDiscovery))
+	// Redirect the original React UI's path (under "/new") to its new path at the root.
+	router.Get("/new/*path", func(w http.ResponseWriter, r *http.Request) {
+		p := route.Param(r.Context(), "path")
+		http.Redirect(w, r, path.Join(o.ExternalURL.Path, strings.TrimPrefix(p, "/new"))+"?"+r.URL.RawQuery, http.StatusFound)
+	})
 
+	router.Get("/classic/alerts", readyf(h.alerts))
+	router.Get("/classic/graph", readyf(h.graph))
+	router.Get("/classic/status", readyf(h.status))
+	router.Get("/classic/flags", readyf(h.flags))
+	router.Get("/classic/config", readyf(h.serveConfig))
+	router.Get("/classic/rules", readyf(h.rules))
+	router.Get("/classic/targets", readyf(h.targets))
+	router.Get("/classic/service-discovery", readyf(h.serviceDiscovery))
+	router.Get("/classic/static/*filepath", func(w http.ResponseWriter, r *http.Request) {
+		r.URL.Path = path.Join("/static", route.Param(r.Context(), "filepath"))
+		fs := server.StaticFileServer(ui.Assets)
+		fs.ServeHTTP(w, r)
+	})
+	// Make sure that "<path-prefix>/classic" is redirected to "<path-prefix>/classic/" and
+	// not just the naked "/classic/", which would be the default behavior of the router
+	// with the "RedirectTrailingSlash" option (https://godoc.org/github.com/julienschmidt/httprouter#Router.RedirectTrailingSlash),
+	// and which breaks users with a --web.route-prefix that deviates from the path derived
+	// from the external URL.
+	// See https://github.com/prometheus/prometheus/issues/6163#issuecomment-553855129.
+	router.Get("/classic", func(w http.ResponseWriter, r *http.Request) {
+		http.Redirect(w, r, path.Join(o.ExternalURL.Path, "classic")+"/", http.StatusFound)
+	})
+
+	router.Get("/version", h.version)
 	router.Get("/metrics", promhttp.Handler().ServeHTTP)
 
 	router.Get("/federate", readyf(httputil.CompressionHandler{
@@ -358,52 +379,43 @@ func New(logger log.Logger, o *Options) *Handler {
 
 	router.Get("/consoles/*filepath", readyf(h.consoles))
 
-	router.Get("/static/*filepath", func(w http.ResponseWriter, r *http.Request) {
-		r.URL.Path = path.Join("/static", route.Param(r.Context(), "filepath"))
-		fs := server.StaticFileServer(ui.Assets)
-		fs.ServeHTTP(w, r)
-	})
-
-	// Make sure that "<path-prefix>/new" is redirected to "<path-prefix>/new/" and
-	// not just the naked "/new/", which would be the default behavior of the router
-	// with the "RedirectTrailingSlash" option (https://godoc.org/github.com/julienschmidt/httprouter#Router.RedirectTrailingSlash),
-	// and which breaks users with a --web.route-prefix that deviates from the path derived
-	// from the external URL.
-	router.Get("/new", func(w http.ResponseWriter, r *http.Request) {
-		http.Redirect(w, r, path.Join(o.ExternalURL.Path, "new")+"/", http.StatusFound)
-	})
-
-	router.Get("/new/*filepath", func(w http.ResponseWriter, r *http.Request) {
-		p := route.Param(r.Context(), "filepath")
-
-		// For paths that the React/Reach router handles, we want to serve the
-		// index.html, but with replaced path prefix placeholder.
-		for _, rp := range reactRouterPaths {
-			if p != rp {
-				continue
-			}
-
-			f, err := ui.Assets.Open("/static/react/index.html")
-			if err != nil {
-				w.WriteHeader(http.StatusInternalServerError)
-				fmt.Fprintf(w, "Error opening React index.html: %v", err)
-				return
-			}
-			idx, err := ioutil.ReadAll(f)
-			if err != nil {
-				w.WriteHeader(http.StatusInternalServerError)
-				fmt.Fprintf(w, "Error reading React index.html: %v", err)
-				return
-			}
-			replacedIdx := bytes.ReplaceAll(idx, []byte("PATH_PREFIX_PLACEHOLDER"), []byte(o.ExternalURL.Path))
-			replacedIdx = bytes.ReplaceAll(replacedIdx, []byte("CONSOLES_LINK_PLACEHOLDER"), []byte(h.consolesPath()))
-			replacedIdx = bytes.ReplaceAll(replacedIdx, []byte("TITLE_PLACEHOLDER"), []byte(h.options.PageTitle))
-			w.Write(replacedIdx)
+	serveReactApp := func(w http.ResponseWriter, r *http.Request) {
+		f, err := ui.Assets.Open("/static/react/index.html")
+		if err != nil {
+			w.WriteHeader(http.StatusInternalServerError)
+			fmt.Fprintf(w, "Error opening React index.html: %v", err)
 			return
 		}
+		idx, err := ioutil.ReadAll(f)
+		if err != nil {
+			w.WriteHeader(http.StatusInternalServerError)
+			fmt.Fprintf(w, "Error reading React index.html: %v", err)
+			return
+		}
+		replacedIdx := bytes.ReplaceAll(idx, []byte("CONSOLES_LINK_PLACEHOLDER"), []byte(h.consolesPath()))
+		replacedIdx = bytes.ReplaceAll(replacedIdx, []byte("TITLE_PLACEHOLDER"), []byte(h.options.PageTitle))
+		w.Write(replacedIdx)
+	}
 
-		// For all other paths, serve auxiliary assets.
-		r.URL.Path = path.Join("/static/react/", p)
+	// Serve the React app.
+	for _, p := range reactRouterPaths {
+		router.Get(p, serveReactApp)
+	}
+
+	// The favicon and manifest are bundled as part of the React app, but we want to serve
+	// them on the root.
+	for _, p := range []string{"/favicon.ico", "/manifest.json"} {
+		assetPath := "/static/react" + p
+		router.Get(p, func(w http.ResponseWriter, r *http.Request) {
+			r.URL.Path = assetPath
+			fs := server.StaticFileServer(ui.Assets)
+			fs.ServeHTTP(w, r)
+		})
+	}
+
+	// Static files required by the React app.
+	router.Get("/static/*filepath", func(w http.ResponseWriter, r *http.Request) {
+		r.URL.Path = path.Join("/static/react/static", route.Param(r.Context(), "filepath"))
 		fs := server.StaticFileServer(ui.Assets)
 		fs.ServeHTTP(w, r)
 	})
@@ -503,11 +515,6 @@ func (h *Handler) testReady(f http.HandlerFunc) http.HandlerFunc {
 	}
 }
 
-// Checks if server is ready, calls f if it is, returns 503 if it is not.
-func (h *Handler) testReadyHandler(f http.Handler) http.HandlerFunc {
-	return h.testReady(f.ServeHTTP)
-}
-
 // Quit returns the receive-only quit channel.
 func (h *Handler) Quit() <-chan struct{} {
 	return h.quitCh
@@ -518,13 +525,13 @@ func (h *Handler) Reload() <-chan chan error {
 	return h.reloadCh
 }
 
-// Run serves the HTTP endpoints.
-func (h *Handler) Run(ctx context.Context) error {
+// Listener creates the TCP listener for web requests.
+func (h *Handler) Listener() (net.Listener, error) {
 	level.Info(h.logger).Log("msg", "Start listening for connections", "address", h.options.ListenAddress)
 
 	listener, err := net.Listen("tcp", h.options.ListenAddress)
 	if err != nil {
-		return err
+		return listener, err
 	}
 	listener = netutil.LimitListener(listener, h.options.MaxConnections)
 
@@ -533,31 +540,18 @@ func (h *Handler) Run(ctx context.Context) error {
 		conntrack.TrackWithName("http"),
 		conntrack.TrackWithTracing())
 
-	var (
-		m = cmux.New(listener)
-		// See https://github.com/grpc/grpc-go/issues/2636 for why we need to use MatchWithWriters().
-		grpcl   = m.MatchWithWriters(cmux.HTTP2MatchHeaderFieldSendSettings("content-type", "application/grpc"))
-		httpl   = m.Match(cmux.HTTP1Fast())
-		grpcSrv = grpc.NewServer()
-	)
+	return listener, nil
+}
 
-	// Prevent open connections to block the shutdown of the handler.
-	m.SetReadTimeout(h.options.ReadTimeout)
-
-	av2 := api_v2.New(
-		h.options.LocalStorage,
-		h.options.TSDBDir,
-		h.options.EnableAdminAPI,
-	)
-	av2.RegisterGRPC(grpcSrv)
-
-	hh, err := av2.HTTPHandler(ctx, h.options.ListenAddress)
-	if err != nil {
-		return err
+// Run serves the HTTP endpoints.
+func (h *Handler) Run(ctx context.Context, listener net.Listener, httpsConfig string) error {
+	if listener == nil {
+		var err error
+		listener, err = h.Listener()
+		if err != nil {
+			return err
+		}
 	}
-
-	hhFunc := h.testReadyHandler(hh)
-
 	operationName := nethttp.OperationNameFunc(func(r *http.Request) string {
 		return fmt.Sprintf("%s %s", r.Method, r.URL.Path)
 	})
@@ -576,13 +570,6 @@ func (h *Handler) Run(ctx context.Context) error {
 
 	mux.Handle(apiPath+"/v1/", http.StripPrefix(apiPath+"/v1", av1))
 
-	mux.Handle(apiPath+"/", http.StripPrefix(apiPath,
-		http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
-			httputil.SetCORS(w, h.options.CORSOrigin, r)
-			hhFunc(w, r)
-		}),
-	))
-
 	errlog := stdlog.New(log.NewStdlibAdapter(level.Error(h.logger)), "", 0)
 
 	httpSrv := &http.Server{
@@ -593,13 +580,7 @@ func (h *Handler) Run(ctx context.Context) error {
 
 	errCh := make(chan error)
 	go func() {
-		errCh <- httpSrv.Serve(httpl)
-	}()
-	go func() {
-		errCh <- grpcSrv.Serve(grpcl)
-	}()
-	go func() {
-		errCh <- m.Serve()
+		errCh <- https.Serve(listener, httpSrv, httpsConfig, h.logger)
 	}()
 
 	select {
@@ -607,27 +588,10 @@ func (h *Handler) Run(ctx context.Context) error {
 		return e
 	case <-ctx.Done():
 		httpSrv.Shutdown(ctx)
-		stopGRPCSrv(grpcSrv)
 		return nil
 	}
 }
 
-// stopGRPCSrv stops a given GRPC server. An attempt to stop the server
-// gracefully is made first. After 15s, the server to forced to stop.
-func stopGRPCSrv(srv *grpc.Server) {
-	stop := make(chan struct{})
-	go func() {
-		srv.GracefulStop()
-		close(stop)
-	}()
-
-	select {
-	case <-time.After(15 * time.Second):
-		srv.Stop()
-	case <-stop:
-	}
-}
-
 func (h *Handler) alerts(w http.ResponseWriter, r *http.Request) {
 
 	var groups []*rules.Group
@@ -859,10 +823,6 @@ func (h *Handler) runtimeInfo() (api_v1.RuntimeInfo, error) {
 	}
 	for _, mF := range metrics {
 		switch *mF.Name {
-		case "prometheus_tsdb_head_chunks":
-			status.ChunkCount = int64(toFloat64(mF))
-		case "prometheus_tsdb_head_series":
-			status.TimeSeriesCount = int64(toFloat64(mF))
 		case "prometheus_tsdb_wal_corruptions_total":
 			status.CorruptionCount = int64(toFloat64(mF))
 		case "prometheus_config_last_reload_successful":
@@ -972,8 +932,15 @@ func (h *Handler) version(w http.ResponseWriter, r *http.Request) {
 }
 
 func (h *Handler) quit(w http.ResponseWriter, r *http.Request) {
-	fmt.Fprintf(w, "Requesting termination... Goodbye!")
-	close(h.quitCh)
+	var closed bool
+	h.quitOnce.Do(func() {
+		closed = true
+		close(h.quitCh)
+		fmt.Fprintf(w, "Requesting termination... Goodbye!")
+	})
+	if !closed {
+		fmt.Fprintf(w, "Termination already in progress.")
+	}
 }
 
 func (h *Handler) reload(w http.ResponseWriter, r *http.Request) {
